/*
*	This was written by Arun Sagar and is placed into the public domain.
*	The author hereby disclaims copyright to this source code.
*
*	Pay it forward. 
*/

#include <WCore>

#include "RBoard.h"

cReversiBoard::CaptureFlag cReversiBoard::grgCaptureFlags[256L*256];
cReversiBoard::CaptureArray cReversiBoard::grgCaptureArray[81*81];
cReversiBoard::Base3TableEntry cReversiBoard::grgBase3Table[256];
ulong cReversiBoard::grgBitCount[256];

ulong cReversiBoard::grgPieceNormalizerMask[]=
{
	0, //none
	0, //edge
	0, //kPiece1
	cReversiBoard::kPieceNormalizerMask // kPiece2
};

void cReversiBoard::
Clear()
{
	AClearData(this->Board);
}

void cReversiBoard::
Reset()
{
	this->Clear();


	this->MakeMove(kPiece1, 4,3);
	this->MakeMove(kPiece1, 3,4);
	this->MakeMove(kPiece2, 3,3);
	this->MakeMove(kPiece2, 4,4);
}

cReversiBoard::LongLine cReversiBoard::ExtractXY(Position d) const
{
	LongLine rLine=kAllEdges;

#define DSXY(i,sh) rLine = cReversiBoardTransferPiece(this->Board[i], rLine, (sh))

#define DSXY1(d,s)	DSXY(d,s);DSXY(d-1,s-1)
#define DSXY2(d,s)	DSXY(d,s);DSXY(d-1,s-1);DSXY(d-2,s-2)
#define DSXY3(d,s)	DSXY(d,s);DSXY1(d-1,s-1);DSXY(d-3,s-3)
#define DSXY4(d,s)	DSXY(d,s);DSXY2(d-1,s-1);DSXY(d-4,s-4)
#define DSXY5(d,s)	DSXY(d,s);DSXY3(d-1,s-1);DSXY(d-5,s-5)
#define DSXY6(d,s)	DSXY(d,s);DSXY4(d-1,s-1);DSXY(d-6,s-6)
#define DSXY7(d,s)	DSXY(d,s);DSXY5(d-1,s-1);DSXY(d-7,s-7)

	switch(d)
	{
		case 0:		DSXY(0,7);	break;
		case 1:		DSXY1(1,7);	break;
		case 2:		DSXY2(2,7);	break;
		case 3:		DSXY3(3,7);	break;
		case 4:		DSXY4(4,7);	break;
		case 5:		DSXY5(5,7);	break;
		case 6:		DSXY6(6,7);	break;
		case 7:		DSXY7(7,7);	break;
		case 8:		DSXY6(7,6); break;
		case 9:		DSXY5(7,5); break;
		case 10:	DSXY4(7,4); break;
		case 11:	DSXY3(7,3); break;
		case 12:	DSXY2(7,2); break;
		case 13:	DSXY1(7,1); break;
		case 14:	DSXY(7,0);	break;
	}

	return rLine;

#undef DSXY7
#undef DSXY6
#undef DSXY5
#undef DSXY4
#undef DSXY3
#undef DSXY2
#undef DSXY1
#undef DSXY
}

cReversiBoard::LongLine cReversiBoard::ExtractYX(Position yx) const
{
	LongLine rLine = kAllEdges;

#define DSYX(i,sh)	rLine = cReversiBoardTransferPiece(this->Board[i], rLine, (sh))
#define DSYX1(d,s)	DSYX(d,s)
#define DSYX2(d,s)	DSYX1(d,s);DSYX(d-1,s+1)
#define DSYX3(d,s)	DSYX2(d,s);DSYX(d-2,s+2)
#define DSYX4(d,s)	DSYX3(d,s);DSYX(d-3,s+3)
#define DSYX5(d,s)	DSYX4(d,s);DSYX(d-4,s+4)
#define DSYX6(d,s)	DSYX5(d,s);DSYX(d-5,s+5)
#define DSYX7(d,s)	DSYX6(d,s);DSYX(d-6,s+6)
#define DSYX8(d,s)	DSYX7(d,s);DSYX(d-7,s+7)

	switch(yx)
	{
		case 0:		DSYX1(0,0);	break;
		case 1:		DSYX2(1,0);	break;
		case 2:		DSYX3(2,0);	break;
		case 3:		DSYX4(3,0);	break;
		case 4:		DSYX5(4,0);	break;
		case 5:		DSYX6(5,0);	break;
		case 6:		DSYX7(6,0);	break;
		case 7:		DSYX8(7,0);	break;
		case 8:		DSYX7(7,1);	break;
		case 9:		DSYX6(7,2);	break;
		case 10:	DSYX5(7,3);	break;
		case 11:	DSYX4(7,4);	break;
		case 12:	DSYX3(7,5);	break;
		case 13:	DSYX2(7,6);	break;
		case 14:	DSYX1(7,7);	break;
	}

	return rLine;
#undef DSYX8
#undef DSYX7
#undef DSYX6
#undef DSYX5
#undef DSYX4
#undef DSYX3
#undef DSYX2
#undef DSYX1
#undef DSYX
}

inline ulong cReversiBoard::
CountLine(ulong rL)
{
	ulong *rgBitCount = cReversiBoard::grgBitCount;
	ulong i;

	i = rL&AFilledInteger8(kPiece1,kPieceBits);
	i ^= ((i>>1)&rL)<<1;

	return rgBitCount[(uint8)i] + rgBitCount[uint8(i>>8)];
}

ulong cReversiBoard::
CountPieces(Piece p) const 
{
	Position i;
	ulong count=0;

	if (kPieceNone == p)
		count = this->CountPieces(kPiece1) + this->CountPieces(kPiece2);
	else if(kPiece1 == p)
	{
		for(i=0;i<8;i++)
			count += this->CountLine(this->Board[i]);
	}
	else
	{
		for(i=0;i<8;i++)
			count += this->CountLine(this->Board[i] ^ kPieceNormalizerMask);
	}

	return count;
}


long cReversiBoard::
FillCaptureList(Piece p, Position x, Position y, CaptureListData rgcl[])
{
	long nCapture=0;
	CaptureData rcDataX, rcDataY, rcDataXY, rcDataYX;
	int i;


	rcDataYX = cReversiBoard::GetCaptureData(this->ExtractYX(7+y-x) ^ cReversiBoard::grgPieceNormalizerMask[p], x);
	rcDataXY = cReversiBoard::GetCaptureData(this->ExtractXY(x+y) ^ cReversiBoard::grgPieceNormalizerMask[p], x);
	rcDataY = cReversiBoard::GetCaptureData(this->ExtractY(x) ^ cReversiBoard::grgPieceNormalizerMask[p], y);
	rcDataX = cReversiBoard::GetCaptureData(this->Board[y] ^ cReversiBoard::grgPieceNormalizerMask[p], x);

#define FILLCAPTURE(rcd, xExp, yExp)\
	for(i=0; i<(rcd); i++)\
	{\
		rgcl[nCapture][0] = x ##xExp;\
		rgcl[nCapture][1] = y ##yExp;\
		nCapture++;\
	}
#define iINC	(i+1)

	FILLCAPTURE(rcDataX.left, -iINC, ;);
	FILLCAPTURE(rcDataX.right, +iINC, ;);
	FILLCAPTURE(rcDataY.left, ;, -iINC);
	FILLCAPTURE(rcDataY.right, ;, +iINC);
	FILLCAPTURE(rcDataXY.left, -iINC, +iINC);
	FILLCAPTURE(rcDataXY.right, +iINC, -iINC);
	FILLCAPTURE(rcDataYX.left, -iINC, -iINC);
	FILLCAPTURE(rcDataYX.right, +iINC, +iINC);

#undef iINC
#undef FILLCAPTURE

	return nCapture;
}




void cReversiBoard::
StuffColumn(LongLine rLine, Position column)
{
	rLine <<= kPieceBits*(7-column);

#define COLSTFD(i)\
		this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], 7-(column));\
		rLine >>= kPieceBits;

	COLSTFD(7);COLSTFD(6);COLSTFD(5);COLSTFD(4);COLSTFD(3);COLSTFD(2);COLSTFD(1);COLSTFD(0);
#undef COLSTFD
}

void cReversiBoard::
StuffXY(LongLine rLine, Position xy)
{
#define DSXY(i,sh) this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));

#define DSXY1(d,s)	DSXY(d,s);DSXY(d-1,s-1)
#define DSXY2(d,s)	DSXY(d,s);DSXY(d-1,s-1);DSXY(d-2,s-2)
#define DSXY3(d,s)	DSXY(d,s);DSXY1(d-1,s-1);DSXY(d-3,s-3)
#define DSXY4(d,s)	DSXY(d,s);DSXY2(d-1,s-1);DSXY(d-4,s-4)
#define DSXY5(d,s)	DSXY(d,s);DSXY3(d-1,s-1);DSXY(d-5,s-5)
#define DSXY6(d,s)	DSXY(d,s);DSXY4(d-1,s-1);DSXY(d-6,s-6)
#define DSXY7(d,s)	DSXY(d,s);DSXY5(d-1,s-1);DSXY(d-7,s-7)

	switch(xy)
	{
		case 0:		DSXY(0,7);	break;
		case 1:		DSXY1(1,7);	break;
		case 2:		DSXY2(2,7); break;
		case 3:		DSXY3(3,7); break;
		case 4:		DSXY4(4,7); break;
		case 5:		DSXY5(5,7); break;
		case 6:		DSXY6(6,7); break;
		case 7:		DSXY7(7,7); break;
		case 8:		DSXY6(7,6); break;
		case 9:		DSXY5(7,5); break;
		case 10:	DSXY4(7,4); break;
		case 11:	DSXY3(7,3); break;
		case 12:	DSXY2(7,2); break;
		case 13:	DSXY1(7,1); break;
		case 14:	DSXY(7,0); break;
	}

#undef DSXY7
#undef DSXY6
#undef DSXY5
#undef DSXY4
#undef DSXY3
#undef DSXY2
#undef DSXY1
#undef DSXY
}


void cReversiBoard::
StuffYX(LongLine rLine, Position yx)
{
#define DSXY(i,sh)	this->Board[i] = cReversiBoardTransferPiece(rLine, this->Board[i], (sh));

#define DSXY1(d,s)	DSXY(d,s)
#define DSXY2(d,s)	DSXY1(d,s);DSXY(d-1,s+1)
#define DSXY3(d,s)	DSXY2(d,s);DSXY(d-2,s+2)
#define DSXY4(d,s)	DSXY3(d,s);DSXY(d-3,s+3)
#define DSXY5(d,s)	DSXY4(d,s);DSXY(d-4,s+4)
#define DSXY6(d,s)	DSXY5(d,s);DSXY(d-5,s+5)
#define DSXY7(d,s)	DSXY6(d,s);DSXY(d-6,s+6)
#define DSXY8(d,s)	DSXY7(d,s);DSXY(d-7,s+7)
	switch(yx)
	{
		case 0:		DSXY1(0,0);	break;
		case 1:		DSXY2(1,0);	break;
		case 2:		DSXY3(2,0);	break;
		case 3:		DSXY4(3,0);	break;
		case 4:		DSXY5(4,0);	break;
		case 5:		DSXY6(5,0);	break;
		case 6:		DSXY7(6,0);	break;
		case 7:		DSXY8(7,0);	break;
		case 8:		DSXY7(7,1);	break;
		case 9:		DSXY6(7,2);	break;
		case 10:	DSXY5(7,3);	break;
		case 11:	DSXY4(7,4);	break;
		case 12:	DSXY3(7,5);	break;
		case 13:	DSXY2(7,6);	break;
		case 14:	DSXY1(7,7);	break;
	}

#undef DSXY8
#undef DSXY7
#undef DSXY6
#undef DSXY5
#undef DSXY4
#undef DSXY3
#undef DSXY2
#undef DSXY1
#undef DSXY
}


void cReversiBoard::
MakeMove(Piece p, Position x, Position y)
{
	LongLine rLine;
	ulong rlPieceNormalizerMask = cReversiBoard::grgPieceNormalizerMask[p];

	Assert((p==kPiece1)||(p==kPiece2));
	Assert(x<8 && y<8);

	// Y-X Diagonal Scan
	rLine = this->ExtractYX(7+y-x) ^ rlPieceNormalizerMask;
	if(cReversiBoard::IsCapturePosition(rLine, x))
		this->StuffYX(cReversiBoard::CaptureLine(rLine, x) ^ rlPieceNormalizerMask, 7+y-x);

	// X+Y Diagonal Scan
	rLine = this->ExtractXY(x+y) ^ rlPieceNormalizerMask;
	if(cReversiBoard::IsCapturePosition(rLine, x))
		this->StuffXY(cReversiBoard::CaptureLine(rLine, x) ^ rlPieceNormalizerMask, x+y);

	// Vertical Scan
	rLine = this->ExtractY(x) ^ rlPieceNormalizerMask;
	if (cReversiBoard::IsCapturePosition(rLine, y))
		this->StuffColumn(cReversiBoard::CaptureLine(rLine, y) ^ rlPieceNormalizerMask, x);

	// Straight Scan
	rLine = this->Board[y] ^ rlPieceNormalizerMask;
	if (cReversiBoard::IsCapturePosition(rLine, x))
		rLine = cReversiBoard::CaptureLine(rLine, x);

	// Set Piece
	rLine = (rLine<<(x*kPieceBits)) & ABuildClearMask(kPieceBits,7*kPieceBits) |
			((ulong)kPiece1<<(7*kPieceBits))&ABuildOverlayMask(kPieceBits,7*kPieceBits);

	// Final
	this->Board[y] = (rLine >> (x*kPieceBits))^rlPieceNormalizerMask;
}





bool cReversiBoard::
IsLegalMove(Piece p, Position x, Position y) const
{
	ulong rlPieceNormalizerMask = cReversiBoard::grgPieceNormalizerMask[p];

	//Assert(this->IsPiece(PieceNone, x, y));

	return (x<8&&y<8) && this->IsPiece(kPieceNone, x, y) &&
		( cReversiBoard::IsCapturePosition(this->Board[y] ^ rlPieceNormalizerMask, x)
		|| cReversiBoard::IsCapturePosition(this->ExtractY(x) ^ rlPieceNormalizerMask, y)
		|| cReversiBoard::IsCapturePosition(this->ExtractXY(x+y) ^ rlPieceNormalizerMask, x)
		|| cReversiBoard::IsCapturePosition(this->ExtractYX(7+y-x) ^ rlPieceNormalizerMask, x));
}


bool cReversiBoard::
IsMoveAvailable(Piece p) const
{
	Position x,y;

	Assert((kPiece1 == p) || (kPiece2 == p));

	for (y=0; y<8; y++)
	{
		for(x=0;x<8;x++)
		{
			if (this->IsLegalMove(p, x, y))
				return true;
		}
	}

	return false;
}

